#include <stdlib.h>
#include <stdio.h>

#include "config.hpp"
#include "utility.hpp"
#include "outstreams.hpp"


static const int INPUT_LINE_LEN = 1024;


void readLocalesConfig  ( string* path, Config* c );
void readFootprintConfig( string* path, Config* c );


// there are three places to look for a given
// config file, in this order:

// 1. In the arg directory (user)
// 2. In the tool directory (user)
// 3. In the tool/resources dir (system default)

Config configInternal;
Config configUserGlobal;



Config::~Config()
{
  for( list<Footprint*>::iterator itr = allFootprints.begin();
       itr != allFootprints.end();
       ++itr )
  {
    delete *itr;
  }

  for( map< string, map<string, Footprint*>* >::iterator itr = type2name2foot.begin();
       itr != type2name2foot.end();
       ++itr )
  {
    delete itr->second;
  }
}








// initialze some data that supports
// the reading of config files
static list<string> validLocales;


void initConfigReading()
{
  validLocales.push_back( "enUS" );
  validLocales.push_back( "enGB" );
  validLocales.push_back( "deDE" );
  validLocales.push_back( "frFR" );
  validLocales.push_back( "ruRU" );
  validLocales.push_back( "itIT" );
  validLocales.push_back( "esMX" );
  validLocales.push_back( "esES" );
  validLocales.push_back( "plPL" );
  validLocales.push_back( "ptBR" );
  validLocales.push_back( "koKR" );
  validLocales.push_back( "zhCN" );
  validLocales.push_back( "zhTW" );
}


void createInternalConfig( Config* c )
{
  for( list<string>::iterator itr = validLocales.begin();
       itr != validLocales.end();
       ++itr )
  {
    c->localePreferences.push_back( *itr );
  }

  c->colors["footertxtfg"] = Color( 1.0f, 1.0f, 1.0f );
  c->colors["footertxtbg"] = Color( 0.0f, 0.0f, 0.0f );
}


void readConfigFiles( string* path, Config* c )
{
  if( path == NULL )
  {
    createInternalConfig( c );
    return;
  }

  readLocalesConfig  ( path, c );
  readFootprintConfig( path, c );
}


// all config files should call this on each input
// line before parsing anything because it will:
// 1) remove comments
// 2) convert white space to plain spaces
void preprocess( string* line )
{
  string::size_type i;

  // 1)
  i = line->find_first_of( "#" );
  if( i != string::npos )
  {
    line->erase( i, string::npos );
  }

  // 2)
  replaceChars( "\t\r\n", " ", line );
}



void readLocalesConfig( string* path, Config* c )
{
  string configFilename( *path );
  configFilename.append( "locales.config" );

  FILE* file = fopen( configFilename.data(), "r" );
  if( file == NULL )
  {
    // if a config file isn't here, no problem just move on
    return;
  }

  int lineNum = 0;
  while( !feof( file ) )
  {
    ++lineNum;

    char lineRaw[INPUT_LINE_LEN];

    char* ret = fgets( lineRaw, INPUT_LINE_LEN, file );
    if( ret == NULL )
    {
      continue;
    }

    string line( lineRaw );
    preprocess( &line );

    strncpy( lineRaw, line.data(), INPUT_LINE_LEN );

    char delims[] = " ";

    char* token = NULL;
    token = strtok( lineRaw, delims );

    if( token == NULL )
    {
      // empty line, no problem
      continue;
    }

    // is token in the list of valid locales?
    bool tokenIsValidLocale = false;

    for( list<string>::iterator validItr = validLocales.begin();
         validItr != validLocales.end();
         ++validItr )
    {
      if( strncmp( token, (*validItr).data(), INPUT_LINE_LEN ) == 0 )
      {
        tokenIsValidLocale = true;

        // found one! If this locale isn't already in the preferences
        // order for the config object, then add it to the back
        bool alreadyInPrefs = false;
        for( list<string>::iterator prefItr = c->localePreferences.begin();
             prefItr != c->localePreferences.end();
             ++prefItr )
        {
          if( *prefItr == *validItr )
          {
            alreadyInPrefs = true;
            break;
          }
        }

        if( !alreadyInPrefs )
        {
          c->localePreferences.push_back( *validItr );
        }

        break;
      }
    }

    if( !tokenIsValidLocale )
    {
      printWarning( "Invalid locale '%s' in %s:%d, ignoring.\n",
                    token,
                    configFilename.data(),
                    lineNum );
      continue;
    }

    token = strtok( NULL, delims );
    if( token != NULL )
    {
      printWarning( "Invalid extra token '%s' in %s:%d, ignoring.\n",
                    token,
                    configFilename.data(),
                    lineNum );
      continue;
    }
  }
}



enum readFootprintMode
{
  RFM_OBJNAMES,
  RFM_CELLS,
};


void readFootprintConfig( string* path, Config* c )
{
  string configFilename( *path );
  configFilename.append( "footprints.config" );

  FILE* file = fopen( configFilename.data(), "r" );
  if( file == NULL )
  {
    // if a config file isn't here, no problem just move on
    return;
  }


  readFootprintMode mode = RFM_OBJNAMES;

  list<string> units;
  list<string> doodads;
  list<string> destructs;
  list<string> resources;
  list<string> nobuildmains;

  Footprint* footprint = NULL;


  int lineNum = 0;
  while( !feof( file ) )
  {
    ++lineNum;

    char lineRaw[INPUT_LINE_LEN];

    char* ret = fgets( lineRaw, INPUT_LINE_LEN, file );
    if( ret == NULL )
    {
      continue;
    }

    string line( lineRaw );
    preprocess( &line );

    strncpy( lineRaw, line.data(), INPUT_LINE_LEN );

    // declare vars for two parsing modes below
    char* delims;
    char* token;
    string type;
    string name;


    switch( mode )
    {
      case RFM_OBJNAMES:

        delims = " ";
        token = strtok( lineRaw, delims );

        if( token == NULL )
        {
          // empty line, no problem
          continue;
        }

        type.assign( token );


        // check if objname mode is over
        if( type == "{" )
        {
          mode = RFM_CELLS;

          footprint = new Footprint();


          token = strtok( NULL, delims );
          if( token != NULL )
          {
            printWarning( "Unexpected extra token %s in %s:%d, brace should have its own line.\n",
                          token,
                          configFilename.data(),
                          lineNum );
          }

          continue;
        }


        // otherwise parse objname

        if( type != "unit"     &&
            type != "doodad"   &&
            type != "destruct" &&
            type != "resource" &&
            type != "nobuildmain" )
        {
          printWarning( "Unknown placed object type '%s' in %s:%d, ignoring line.\n",
                        token,
                        configFilename.data(),
                        lineNum );
          continue;
        }

        token = strtok( NULL, delims );
        if( token == NULL )
        {
          printWarning( "Expected object name after %s in %s:%d, ignoring line.\n",
                        type.data(),
                        configFilename.data(),
                        lineNum );
          continue;
        }

        name.assign( token );

        if( type == "unit"        ) { units       .push_back( name ); } else
        if( type == "doodad"      ) { doodads     .push_back( name ); } else
        if( type == "destruct"    ) { destructs   .push_back( name ); } else
        if( type == "resource"    ) { resources   .push_back( name ); } else
        if( type == "nobuildmain" ) { nobuildmains.push_back( name ); }

        token = strtok( NULL, delims );
        if( token != NULL )
        {
          printWarning( "Unexpected extra token %s in %s:%d, ignoring.\n",
                        token,
                        configFilename.data(),
                        lineNum );
          continue;
        }
      break;


      case RFM_CELLS:

        delims = " ,";
        token = strtok( lineRaw, delims );

        if( token == NULL )
        {
          // empty line, no problem
          continue;
        }

        type.assign( token );

        // check if cells mode is over
        if( type == "}" )
        {
          mode = RFM_OBJNAMES;

          token = strtok( NULL, delims );
          if( token != NULL )
          {
            printWarning( "Unexpected extra token %s in %s:%d, brace should have its own line.\n",
                          token,
                          configFilename.data(),
                          lineNum );
          }


          map<string, Footprint*>* units_name2foot = c->type2name2foot["unit"];
          if( units_name2foot == NULL )
          {
            units_name2foot = new map<string, Footprint*>();
            c->type2name2foot["unit"] = units_name2foot;
          }

          for( list<string>::iterator uItr = units.begin();
               uItr != units.end();
               ++uItr )
          {
            units_name2foot->insert( make_pair( *uItr, footprint ) );
          }


          map<string, Footprint*>* doodads_name2foot = c->type2name2foot["doodad"];
          if( doodads_name2foot == NULL )
          {
            doodads_name2foot = new map<string, Footprint*>();
            c->type2name2foot["doodad"] = doodads_name2foot;
          }

          for( list<string>::iterator dItr = doodads.begin();
               dItr != doodads.end();
               ++dItr )
          {
            doodads_name2foot->insert( make_pair( *dItr, footprint ) );
          }


          map<string, Footprint*>* destructs_name2foot = c->type2name2foot["destruct"];
          if( destructs_name2foot == NULL )
          {
            destructs_name2foot = new map<string, Footprint*>();
            c->type2name2foot["destruct"] = destructs_name2foot;
          }

          for( list<string>::iterator dItr = destructs.begin();
               dItr != destructs.end();
               ++dItr )
          {
            destructs_name2foot->insert( make_pair( *dItr, footprint ) );
          }


          map<string, Footprint*>* resources_name2foot = c->type2name2foot["resource"];
          if( resources_name2foot == NULL )
          {
            resources_name2foot = new map<string, Footprint*>();
            c->type2name2foot["resource"] = resources_name2foot;
          }

          for( list<string>::iterator dItr = resources.begin();
               dItr != resources.end();
               ++dItr )
          {
            resources_name2foot->insert( make_pair( *dItr, footprint ) );
          }


          map<string, Footprint*>* nobuildmains_name2foot = c->type2name2foot["nobuildmain"];
          if( nobuildmains_name2foot == NULL )
          {
            nobuildmains_name2foot = new map<string, Footprint*>();
            c->type2name2foot["nobuildmain"] = nobuildmains_name2foot;
          }

          for( list<string>::iterator dItr = nobuildmains.begin();
               dItr != nobuildmains.end();
               ++dItr )
          {
            nobuildmains_name2foot->insert( make_pair( *dItr, footprint ) );
          }

          units       .clear();
          doodads     .clear();
          destructs   .clear();
          resources   .clear();
          nobuildmains.clear();

          footprint = NULL;

          continue;
        }



        // otherwise parse relative cell coordinates
        int rcx = atoi( token );

        token = strtok( NULL, delims );
        if( token == NULL )
        {
          printWarning( "Expected a second coordinate in %s:%d, ignoring line.\n",
                        configFilename.data(),
                        lineNum );
          continue;
        }

        int rcy = atoi( token );

        footprint->relativeCoordinates.push_back( rcx );
        footprint->relativeCoordinates.push_back( rcy );


        token = strtok( NULL, delims );
        if( token != NULL )
        {
          printWarning( "Unexpected extra token %s in %s:%d, ignoring.\n",
                        token,
                        configFilename.data(),
                        lineNum );
        }
      break;


      default:
        printError( "Unknown mode while parsing footprint.config!" );
        exit( -1 );
    }
  }


  // if we end mid-footprint parse, gotta clean it up, otherwise
  // its in the master list for the Config and it will be cleaned
  if( footprint != NULL )
  {
    delete footprint;
  }
}
